#ifdef WIN32
#define _CRT_SECURE_NO_WARNINGS
#endif

// #define DEBUG_HASH

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef WIN32
#include <io.h>
#include <direct.h>

#define open _open
#define read _read
#define close _close
#define O_RDONLY _O_RDONLY
#else
#include <unistd.h>
#include <fcntl.h>
#endif
#include <stdlib.h>

#include "util.h"
#include "xmalloc.h"
#include "object.h"
#include "label.h"
#include "location.h"
#include "dwarf.h"
#include "file.h"
#include "fileref.h"
#include "code.h"
#include "function.h"
#include "directory.h"
#include "source.h"

#ifdef WIN32
#define read _read
#endif

#define HASH_SIZE 0x100000

#ifdef DEBUG_HASH
static codeT	*hash_table[HASH_SIZE] = {0};
#endif

static codeT	**hash_tailPP[HASH_SIZE] = {0};

static void
hash_assembler(codeT *codeP)
{
	char		*P;
	int			hash;
	int			c;
	codeT		**hashPP;

	hash = 0;
	P    = codeP->m_codeP;
	if (P) {
		if (*P == 'j') {
			for (; (c = *P) && c != ' '; ++P) {
				hash <<= 1;
				hash  ^= *P;
			}
		} else {
			for (; (c = *P); ++P) {
				hash <<= 1;
				hash  ^= *P;
		}	}
		hash &= (HASH_SIZE - 1);
	}
	
	if ((hashPP = hash_tailPP[hash])) {
		*hashPP = codeP; 
	}
#ifdef DEBUG_HASH
  	  else {
		hash_table[hash] = codeP;
	}
#endif

	hash_tailPP[hash] = &(codeP->m_hashP);
}

#ifdef DEBUG_HASH
static void
dump_hash(void)
{
	codeT *codeP;
	int i;

	for (i = 0; i < HASH_SIZE; ++i) {
		if ((codeP = hash_table[i])) {
			fprintf(stderr, "%d: ", i);
			for (; codeP; codeP = codeP->m_hashP) {
				fprintf(stderr, "%s/", codeP->m_codeP);
			}
			fprintf(stderr, "\n");
	}	}
	fprintf(stderr, "End of hash\n");
}
#endif

Csource			*Csource::g_headP     = 0;
Csource			*Csource::g_tailP     = 0;

extern	int		g_use_frame_address;
extern	int		g_match_lines;
extern	int		g_max_function_lth;
extern	int		g_read;
extern	unsigned int	g_hypertext_cnt;
extern	FILE 	*taF;
extern	FILE	*tempF;

/* Source is preserved in the order visited */

Csource::Csource(char *nameP)
{
	m_nextP          = 0;
	m_has_clones     = 0;
	m_assembler_size = 0;
	m_assemblerP     = 0;
	m_functionsP     = 0;
	m_filerefsP      = 0;
	m_fileP          = Cfile::locate(nameP);
	m_hypertext_cnt  = ++g_hypertext_cnt;
	if (taF) {
		m_parentP = Cdirectory::locate(nameP);
	} else {
		m_parentP = 0;
}	}

extern int	g_stdin_lineno;

Cfileref *
Csource::locate_fileref(int number) const
{
	Cfileref	*filerefP;

	for (filerefP = m_filerefsP; filerefP; filerefP = filerefP->m_nextP) {
		if (filerefP->m_number == number) {
			break;
	}	}
	return(filerefP);
}

static labelT	*g_labelsP   = 0;

static labelT *
find_label(const char *stringP)
{
	labelT	*labelP;
	const char *P;

	for (labelP = g_labelsP; (P = labelP->labelP); ++labelP) {
		if (!strcmp(P, stringP)) {
			return(labelP);
	}	}
	return(0);
}

int
Csource::load(char *assemblerP)
{
	static	int		g_labels_cnt = 0;

	int			fileno;
	struct stat buf;
	char		*atP, *startP, *endP;
	int			labels_seen, codes_seen;
	codeT		*codeP, *end_codesP;
	labelT		*labelP, *string_labelP;
	int			lth, lth1, lth2, segment, source_lineno;
	char		*debug_abbrevP, *debug_infoP, *debug_strP;
	int			ret;

	Cfunction	*functionP, **functionsPP;
	Cfileref	*filerefP;
	char		*nameP, *P, *P1, *P2, *absolute_pathP;
	int			number;
	char		c, c1, quote;

	Cdwarf		dwarf;
	int			locations0, locations, max_locations;
	locationT	*locationsP, *locationP;

	Cfile		*fileP, *last_fileP;
	int			last_lineno;


	nameP  = m_fileP->m_nameP;
	fileno = open(assemblerP, O_RDONLY, 0);
	if (fileno < 0) {
		fprintf(stderr, "Line %d: Can't open %s\n", g_stdin_lineno, assemblerP);
		perror("");
		return -1;
	}
	// We have to scan the input more than once so read the entire input
	// into memory

	ret = fstat(fileno, &buf);
	if (ret != 0) {
		fprintf(stderr,"** Line %d: fstat %s returned %d\n", g_stdin_lineno, nameP, ret);
		perror("");
		goto bad_source;
	}

	// Allocate a large enough memory buffer

	lth = buf.st_size;
	if (!lth) {
		fprintf(stderr, "** Line %d: %s is empty\n", g_stdin_lineno, nameP);
		goto bad_source;
	}

	m_assemblerP     = atP = (char *) Xmalloc(lth + 2);
	m_assembler_size = lth;
	
	// Read the assembler into memory

	for (endP = atP + lth; atP < endP; atP += ret) {
		segment = endP - atP;
		if (segment > SSIZE_MAX) {
			segment = SSIZE_MAX;
		}
		ret = read(fileno, atP, segment);
		if (ret <= 0) {
			fprintf(stderr, "** Line %d: read %s returned %d\n", g_stdin_lineno, nameP, ret);
			perror("");
bad_source:
			close(fileno);
			return -1;
	}	}
	assert(atP == endP);
	if (atP[-1] != '\n') {
		*endP++ = '\n';
	}
	*endP = 0;

	close(fileno);

	// Remove comments from assembler
	// Remove needless blanks

	++g_read;
	quote = 0;
	for (atP = P = startP = m_assemblerP; (c = *atP); ++atP) {
		if (quote) {
			switch (c) {
			case 0:
				fprintf(stderr, "** Line %d: unterminated quote in %s\n", g_stdin_lineno, nameP);
				return -1;
			case '\n':
				fprintf(stderr, "** Line %d: quoted newline character in %s\n", g_stdin_lineno, nameP);
				return -1;
			case '\'':
			case '"':
				if (quote == c) {
					quote = 0;
				}
				break;
			case '\\':
				c = *(++atP);
				switch (c) {
				case 0:
				case '\n':
					// Ignore the backslash
					--atP;
					continue;
				}
				*P++ = '\\';
				break;
			}
		} else {
			switch (c) {
			case '\r':
				if (atP[1] == '\n') {
					continue;
				}
				c = '\n';
				goto end_of_line;
			case ';':
				c = '\n';
			case '\n':
end_of_line:
				if (P > startP && P[-1] == ' ') {
					--P;
				}
				if (P > startP && P[-1] == '\n') {
					continue;
				}
				break;
			case '\t':
				c = ' ';
			case ' ':
				if (P > startP && ((c1 = P[-1]) == ' ' || c1 == ',')) {
					continue;
				}
				break;
			case ',':
				if (P > startP && P[-1] == ' ') {
					--P;
				}
				break;
			case '\\':
				c = *(++atP);
				switch (c) {
				case 0:
				case '\n':
					// Ignore the backslash
					--atP;
					continue;
				}
				*P++ = '\\';
				break;
			case '#':
				while (*(++atP) != '\n' );
				--atP;
				continue;
			case '/':
				if (atP[1] == '*') {
					for (atP += 2; ;) {
						switch (*atP) {
						case 0:
							atP  -= 2;
							break;
						case '*':
							if (atP[1] == '/') {
								++atP;
								break;
							}
						default:
							continue;
						}
						break;
					}
					continue;
				}
				break;
			case '"':
			case '\'':
				quote = c;
		}	}
		*P++ = c;
	}	
	*P = 0;
	m_assembler_size = P - startP;

	if (!g_headP) {
		g_headP = this;
	} else {
		g_tailP->m_nextP = this;
	}
	g_tailP = this;

	functionP   = 0;
	functionsPP = &m_functionsP;

	labels_seen = 0;
	codes_seen  = 0;

	// Now count labels, jumps and assembler lines

	for (atP = startP; (c = *atP); atP = strchr(atP, '\n')+1) {
		// Now process this line

		for (; *atP == ' '; ++atP);
		for (P = atP;; ++P) {
			c = *P;
			if (!label_char[(int) c]) {
				if (c == ':') {
					++labels_seen;
					atP = P + 1;
				}
				break;
		}	}
		for (P = atP; *P == ' '; ++P);
		switch (*P) {
		case 0:
			assert(0);
			continue;
		case '.':
		case '\n':
			continue;
		default:
			++codes_seen;
	}	}
	
	// At most one label for each jump
	m_codesP     = codeP = (codeT *) Xmalloc(codes_seen * sizeof(codeT));
	m_codes      = codes_seen;
	end_codesP   = codeP + codes_seen;

	if (g_labels_cnt < labels_seen) {
		Xfree(g_labelsP);
		g_labelsP = (labelT *) Xmalloc((labels_seen + 1) * sizeof(labelT));
		g_labels_cnt = labels_seen;
	}

	labelP        = g_labelsP;
	filerefP      = 0;
	source_lineno = 0;
	debug_infoP   = 0;
	debug_abbrevP = 0;
	debug_strP    = 0;
	string_labelP = 0;
	for (atP = startP; (c = *atP); ) {
		// Now process this line

		for (; *atP == ' '; ++atP);
		for (P = atP;; ++P) {
			c = *P;
			if (!label_char[(int) c]) {
				if (c == ':') {
					labelP->labelP  = atP;
					*P = 0;
					labelP->lineno  = codeP - m_codesP;
					string_labelP   = labelP;
					labelP->stringP = 0;
					++labelP;
					++P;
					atP = P;
				}
				break;
		}	}
		for (P = atP; *P == ' '; ++P);
		switch (*P) {
		case 0:
			assert(0);
		case '\n':
			atP = P+1;
			continue;
		case '.':
			++P;
			switch (*P) {
			case 's':	//.s
				if (!strncmp(P, "section .", 9)) {
					P += 9;
					if (!strncmp(P, "debug_info\n", 11)) {
						assert(!debug_infoP);
						atP = debug_infoP = P + 11;
						continue;
					}
					if (!strncmp(P, "debug_abbrev\n", 13)) {
						assert(!debug_abbrevP);
						atP = debug_abbrevP = P + 13;
						continue;
					}
					if (!strncmp(P, "debug_str,", 10)) {
						assert(!debug_strP);
						for (P += 10; *P != '\n'; ++P);
						atP = debug_strP = ++P;
						continue;
				}	}
				if (string_labelP && !strncmp(P, "string ", 7)) {
					for (P += 7; *P == ' '; ++P);
					string_labelP->stringP = P;
					if (*P != '"') {
						P = strchr(P,'\n');
					} else { 
						while ((c = *++P) != '"') {
							if (c == '\\') {
								++P;
						}	}
						++P;
					}
					string_labelP->lth = P - string_labelP->stringP;
					string_labelP      = 0;
				}
				break;
			case 't':	// .t
				if (!strncmp(P, "type ", 5)) {		// .type	
					for (P += 5; *P == ' '; ++P);
					P1 = end_symbol_name(P);
					if (!P1) {
						break;
					}
					P2 = strchr(P1, ',');
					if (!P2) {
						break;
					}
					for (++P2; *P2 == ' '; ++P2);
					if (!strncmp(P2, "@function", 9)) {
						c   = *P1;
						*P1 = 0;
						if (functionP) {
							lth = functionP->m_end_code - functionP->m_start_code;
							if (lth > g_max_function_lth) {
								g_max_function_lth = lth;
						}	}
						functionP = new Cfunction(P, this);
						if (!functionP) {
							outofmemory();
						}
						functionP->m_start_code = codeP - m_codesP;
						functionP->m_end_code   = functionP->m_start_code;
						*functionsPP            = functionP;
						functionsPP             = &(functionP->m_nextP);
						*P1 = c;
					}
				}
				break;
			case 'f':	// .f
				if (!strncmp(P, "file ", 5)) {		// .file
					for (P += 5; *P == ' '; ++P);
					number = strtol(P, &P1, 10);
					if (P == P1) {
						break;
					}
					if (*P1 != ' ') {
						break;
					}
					for (++P1; *P1 == ' '; ++P1);
					if (*P1 != '"') {
						break;
					}
					P = P1 + 1;
					P1 = strchr(P, '"');
					if (!P1) {
						break;
					}
					c   = *P1;
					*P1 = 0;
					absolute_pathP = absolute_path(P);
					*P1 = c;
					filerefP = new Cfileref(number, absolute_pathP);
					if (!filerefP) {
						outofmemory();
					}
					// Add to front so later definitions override
					// earlier ones
					filerefP->m_nextP = m_filerefsP;
					m_filerefsP       = filerefP;
				}
				break;
			case 'l':	// .l
				if (!strncmp(P, "loc ", 4)) {		// .loc
					for (P += 4; *P == ' '; ++P);
					number = strtol(P, &P1, 10);
					if (P == P1) {
						break;
					}
					if (!filerefP || filerefP->m_number != number) {
						filerefP = locate_fileref(number);
						if (!filerefP) {
							assert(0);
					}	}
					for (P = P1; *P == ' '; ++P);
					source_lineno = strtol(P, &P1, 10);
					if (functionP && functionP->m_source_lineno < 0 && filerefP && filerefP->m_fileP == m_fileP) {
						functionP->m_source_lineno = source_lineno;
					}
				}
				break;
			}
			break;
		default:
			string_labelP = 0;
			if (!strncmp(P, "ret\n", 4)) {
				if (functionP) {
					functionP->m_end_code = (codeP - m_codesP) + 1;
			}	}

			if (functionP) {
				memset(codeP, 0, sizeof(codeT));
				codeP->m_functionP     = functionP;
				if (filerefP) {
					codeP->m_fileP     = filerefP->m_fileP;
				}
				codeP->m_source_lineno = source_lineno;
				codeP->m_codeP         = P;
				++codeP;
			}

			for (; *P != '\n'; ++P);
			*P = 0;
			atP = P+1;
	
			continue;
		}
		atP = strchr(P, '\n');
		assert(atP);
		++atP;
	}
	if (functionP) {
		lth = functionP->m_end_code - functionP->m_start_code;
		if (lth > g_max_function_lth) {
			g_max_function_lth = lth;
	}	}
	labelP->labelP = 0;
	assert(labelP - g_labelsP == labels_seen);
	m_codes = codeP - m_codesP;
	assert(m_codes <= codes_seen);
	end_codesP  = codeP;

	for (codeP = m_codesP; codeP < end_codesP; ++codeP) {
		atP = codeP->m_codeP;
		if (*atP != 'j') {
			P = strchr(atP, '$');
			if (P && label_char[(int) *++P] > 1) {
				labelP = find_label(P);
				if (labelP && labelP->stringP) {
					lth  = strlen(labelP->labelP);
					lth1 = labelP->lth;
					if (lth1 <= lth) {
						memcpy(P, labelP->stringP, lth1);
						P += lth1;
						if (lth1 < lth) {
							strcpy(P, P + lth - lth1);
						}
					} else {
						segment = strlen(atP);
						codeP->m_codeP = P1 = (char *) Xmalloc(segment + lth1 - lth + 1);
						lth2 = P - atP;
						memcpy(P1, atP, lth2);
						P1 += lth2;
						memcpy(P1, labelP->stringP, lth1);
						P1 += lth1;
						strcpy(P1, P + lth);
			}	}	}		
		} else {
			for (++atP; (c = *atP) && c != ' '; ++atP);
			if (!c) {
				assert(0);
				continue;
			}
			for (++atP; (c = *atP) && c == ' '; ++atP);
			if (!c) {
				assert(0);
				continue;
			}
			labelP = find_label(atP);
			if (labelP) {
				// Make jump relative
				// Otherwise this is something like jmp *%eax
				codeP->m_jumpP       = m_codesP + labelP->lineno;
	}	}	}

	if (!g_use_frame_address && debug_abbrevP && debug_infoP) {
		dwarf.load(g_labelsP, debug_abbrevP, debug_infoP, debug_strP);
		codeP         = m_codesP;
		max_locations = 0;
		// Allow this to be decremented and testing less.
		locationsP    = ((locationT *) 0)+1;
		for (functionP = m_functionsP; functionP; functionP = functionP->m_nextP) {
			P         = functionP->m_signatureP;
			P1        = functionP->m_nameP;
			locationP = 0;
			dwarf.load_locations(P, P1, &locationP, 0 /* Don't load */);
			locations = locations0 = locationP - ((locationT *) 0);
			if (locations != 0) {
				if (max_locations <= locations) {
					if (!max_locations) {
						max_locations = 128;
					}
					while (max_locations <= locations) {
						max_locations <<= 1;
					}
					locationsP = (locationT *) alloca(sizeof(locationT) * max_locations);
					if (!locationsP) {
						outofmemory();
				}	}
				locationP = locationsP;
				dwarf.load_locations(P, P1, &locationP, 1);
				locations = locationP - locationsP;
				assert(locations <= locations0);
			}
			functionP->update_names(locations, locationsP, codeP);
	}	}	

	// Do after any name changing
	if (!g_match_lines) {
		for (codeP = m_codesP; codeP < end_codesP; ++codeP) {
			hash_assembler(codeP);
		}
	} else {
		last_fileP  = 0;
		last_lineno = 0;
		for (codeP = m_codesP; codeP < end_codesP; ++codeP) {
			if ((fileP = codeP->m_fileP)) {
				if (fileP == last_fileP) {
					if (codeP->m_source_lineno == last_lineno) {
						continue;
					}
				} else {
					last_fileP = fileP;
			}	}
			last_lineno = codeP->m_source_lineno;
			hash_assembler(codeP);
	}	}
#ifdef DEBUG_HASH
	dump_hash();
#endif
	return 0;
}

void
Csource::emit_ta(void) const
{
	Cdirectory	*parentP;
	char		*nameP, *P;

	parentP = m_parentP;

	if (parentP) {
		// Making the count 2 causes automatic directory to treat as must
		// not be collapsed
		parentP->has_clones(2);
	}
	fprintf(taF, "$INSTANCE S%p S\n", this);
	if (parentP) {
		fprintf(taF, "contain D%p S%p\n", parentP, this);
	}
	nameP = m_fileP->m_nameP;
	P     = strrchr(nameP, '/');
	if (P && P[1]) {
		nameP = P+1;
	}
	fprintf(tempF, "S%p {label=\"%s\" file=\"./%s\"}\n", this, nameP, nameP);
}

